package com.persagy.apm.diagnose.indicatorrecord.service.impl;

import com.alibaba.fastjson.JSONArray;
import com.alibaba.fastjson.JSONObject;
import com.persagy.apm.diagnose.config.ComputePropertiesConfig;
import com.persagy.apm.diagnose.config.RabbitConfig;
import com.persagy.apm.diagnose.constant.DiAgnoseConst;
import com.persagy.apm.diagnose.constant.EnumDataTimeType;
import com.persagy.apm.diagnose.constant.EnumTimeType;
import com.persagy.apm.diagnose.indicatorcompute.model.ProjectIndicatorCompute;
import com.persagy.apm.diagnose.indicatorcompute.service.IProjectIndicatorComputeService;
import com.persagy.apm.diagnose.indicatorrecord.model.dto.*;
import com.persagy.apm.diagnose.indicatorrecord.service.IMonitorIndicatorRecordService;
import com.persagy.apm.diagnose.service.CenterDataService;
import com.persagy.apm.diagnose.service.EnergyAlarmServiceImpl;
import com.persagy.apm.diagnose.service.dto.AlarmRuleDTO;
import com.persagy.apm.diagnose.service.dto.CalculateMethodDTO;
import com.persagy.apm.diagnose.service.dto.ProjectDTO;
import com.persagy.apm.diagnose.service.dto.TimeDataDTO;
import com.persagy.apm.diagnose.timerecord.model.CalculateTimeRecord;
import com.persagy.apm.diagnose.timerecord.service.ICalculateTimeRecordService;
import com.persagy.apm.diagnose.utils.*;
import com.persagy.apm.energyalarmstarter.alarmdata.service.AlarmConfigServiceImpl;
import com.persagy.framework.ems.data.core.enumeration.EMSOrder;
import com.persagy.framework.ems.data.core.enumeration.SpecialOperator;
import com.persagy.framework.ems.data.mvc.dao.CoreDao;
import com.persagy.framework.ems.data.pojo.hbase.MonitorIndicatorRecord;
import lombok.extern.slf4j.Slf4j;
import org.apache.commons.lang.StringUtils;
import org.springframework.beans.BeanUtils;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.scheduling.annotation.Async;
import org.springframework.scheduling.annotation.AsyncResult;
import org.springframework.stereotype.Service;
import org.springframework.util.CollectionUtils;

import javax.annotation.Resource;
import java.math.BigDecimal;
import java.util.*;
import java.util.concurrent.CountDownLatch;
import java.util.concurrent.Future;
import java.util.stream.Collectors;

/**
 * 监测指标历史记录(MonitorIndicatorRecord) service层
 *
 * @author lixing
 * @version V1.0 2021-09-10 00:13:27
 */
@Slf4j
@Service
public class MonitorIndicatorRecordServiceImpl implements IMonitorIndicatorRecordService {

    @Autowired
    private RedisLock lockUtil;

    @Autowired
    private ICalculateTimeRecordService calculateTimeRecordService;


    @Autowired
    private AlarmConfigServiceImpl alarmConfigService;

    @Autowired
    private RedisUtil redisUtil;

    @Autowired
    private EnergyAlarmServiceImpl energyAlarmService;

    @Autowired
    private ComputePropertiesConfig computePropertiesConfig;


    @Autowired
    private CenterDataService centerDataService;


    public final static String alarmSlipt = "_";


//    private final static Long SendDataTimeKeyTime = 24 * 60 * 60 * 1000L;


    @Autowired
    private IProjectIndicatorComputeService projectIndicatorComputeService;

    @Resource(name = "ZillionDao")
    private CoreDao coredao;
    @Autowired
    private RabbitConfig rabbitConfig;


    /**
     * 查询监测指标历史记录
     *
     * @return List<MonitorIndicatorRecord>
     * @author lixing
     */
    @Override
    public List<MonitorIndicatorRecord> queryMonitorIndicatorRecordList(QueryMonitorIndicatorRecordDTO queryMonitorIndicatorRecordDTO) throws Exception {
        MonitorIndicatorRecord query = new MonitorIndicatorRecord();
        query.setMonitorIndicatorId(queryMonitorIndicatorRecordDTO.getMonitorIndicatorId());
        query.setProject(queryMonitorIndicatorRecordDTO.getProjectId());
        query.setObjId(queryMonitorIndicatorRecordDTO.getObjId());
        query.setSpecialOperation("dataTime", SpecialOperator.$gte, queryMonitorIndicatorRecordDTO.getTimeFrom());
        query.setSpecialOperation("dataTime", SpecialOperator.$lt, queryMonitorIndicatorRecordDTO.getTimeTo());
        return coredao.query(query);
    }


    @Override
    @Async("asyncComputeDataExecutor")
    public Future<String> computeIndicatorData(ProjectDTO projectDTO, ProjectIndicatorCompute projectIndicatorCompute, CountDownLatch latch) throws Exception {
        long time = System.currentTimeMillis() + (20 * 1000);
        try {
            if (!lockUtil.lock(projectDTO.getProjectId() + DiAgnoseConst.RedisConstants.ComputeProjectLockKey, String.valueOf(time))) {
                // 获取不到锁,略过
                return new AsyncResult<>("【监测指标数据计算线程】项目:" + projectDTO.getProjectId() + "被占用");
            }

            long start = System.currentTimeMillis();
            AsyncResult<String> errorMsg = computeProjectIndicatorData(projectDTO, projectIndicatorCompute);
            if (errorMsg != null) {
                return errorMsg;
            }
            long end = System.currentTimeMillis();
            return new AsyncResult<>("【监测指标数据计算线程】计算项目指标数据用时:" + projectDTO.getProjectId() + ";" + (end - start));
        } finally {
            // 释放锁
            lockUtil.unlock(projectDTO.getProjectId() + DiAgnoseConst.RedisConstants.ComputeProjectLockKey, String.valueOf(time));
            latch.countDown();
        }

    }


    private AsyncResult<String> computeProjectIndicatorData(ProjectDTO projectDTO, ProjectIndicatorCompute projectIndicatorCompute) throws Exception {

        {
            //判断当前计算轮次是否大于缓存中最新计算轮次,如果是,则证明已经被计算过,则略过
            Object redisComputeRound = redisUtil.hget(DiAgnoseConst.RedisConstants.TargetComputeTime, projectDTO.getProjectId());
            if (redisComputeRound != null && projectIndicatorCompute != null
                    && projectIndicatorCompute.getComputeRound() < Long.parseLong(redisComputeRound.toString())) {
                return new AsyncResult<>("【监测指标数据计算线程】当前项目:" + projectDTO.getProjectId() + "被占用");
            }
        }
        //通过报警服务查询报警规则
        List<AlarmRuleDTO> projectAlarmRules = energyAlarmService.getProjectAlarmRules(projectDTO.getProjectId());
        if (CollectionUtils.isEmpty(projectAlarmRules)) {
            return new AsyncResult<>("【监测指标数据计算线程】当前项目:" + projectDTO.getProjectId() + "未设置报警规则");
        }

        Set<String> objIdAndIndicatorId = getObjIdAndInfoCodeFromRules(projectAlarmRules);
        if (CollectionUtils.isEmpty(objIdAndIndicatorId)) {
            return new AsyncResult<>("【监测指标数据计算线程】当前项目:" + projectDTO.getProjectId() + "报警规则未关联对象");
        }
        Map<String, Date> objId_indicatorIdAndComputeTIme = getObjId_IndicatorIdAndComputeTIme(projectDTO);
        Map<String, CalculateMethodDTO> methodIdAndMethodDTO = energyAlarmService.queryCalculateMethods();

        for (String objIdIndicatorId : objIdAndIndicatorId) {
            int indexOf = objIdIndicatorId.indexOf("_");
            String objId = objIdIndicatorId.substring(0, indexOf);
            String indicatorId = objIdIndicatorId.substring(indexOf + 1);
            String methodId = energyAlarmService.queryObjIndicatorMethodId(projectDTO.getProjectId(), objId, indicatorId);
            if (StringUtils.isBlank(methodId)) {
                log.error("【监测指标数据计算线程】:项目监控指标未设置公式:" + projectDTO.getProjectId() + ";" + objId + ";" + indicatorId);
                continue;
            }
            CalculateMethodDTO calculateMethodDTO = methodIdAndMethodDTO.get(methodId);
            if (calculateMethodDTO == null) {
                log.error("【监测指标数据计算线程】:项目监控指标设置公式为null:" + projectDTO.getProjectId() + ";" + objId + ";" + indicatorId);
                continue;
            }
            Date lastComputeTime = objId_indicatorIdAndComputeTIme.get(objIdIndicatorId);
            if (lastComputeTime == null) {
                lastComputeTime = getDefultStartTime();
            }
            Date computeEndTime = DateUtils.addDays(lastComputeTime, computePropertiesConfig.getRoundDays());
            String formula = calculateMethodDTO.getFormula();
            Date computeTime = computeIndicatorData(projectDTO, objId, lastComputeTime, computeEndTime, indicatorId, formula);
            saveOrUpdateComputeTime(projectDTO.getProjectId(), objId, indicatorId, computeTime);
        }
        // 更新缓存计算轮次
        updateProjectComputeRound(projectDTO.getProjectId(), projectIndicatorCompute);
        return null;
    }

    /**
     * @title getObjIdAndInfoCodeFromRules
     * @description 通过报警条件获取所有对象及其监控点位,objId_infoCode
     * @author leifeng
     * @param: projectAlarmRules
     * @updateTime 2021/10/26 21:24
     * @return: java.util.Set<java.lang.String>
     * @throws
     */
    private Set<String> getObjIdAndInfoCodeFromRules(List<AlarmRuleDTO> projectAlarmRules) {
        Set<String> objIdAndInfoCode = new HashSet<>();
        for (AlarmRuleDTO projectAlarmRule : projectAlarmRules) {
            List<String> objIds = projectAlarmRule.getObjIds();
            if(CollectionUtils.isEmpty(objIds)){
                continue;
            }
            List<String> infoCodes = projectAlarmRule.getInfoCodes();
            if (CollectionUtils.isEmpty(infoCodes)){
                continue;
            }
            for (String infoCode : infoCodes) {
                for (String objId : objIds) {
                    objIdAndInfoCode.add(objId + "_" + infoCode);
                }
            }
        }
        return objIdAndInfoCode;
    }

    /**
     * @title computeIndicatorData
     * @description 计算指标数据,返回最后计算时间
     * @author leifeng
     * @param: projectDTO
     * @param: objId
     * @param: lastComputeTime
     * @param: computeEndTime
     * @param: indicatorId
     * @param: formula
     * @updateTime 2021/9/25 20:11
     */
    private Date computeIndicatorData(ProjectDTO projectDTO, String objId, Date lastComputeTime, Date computeEndTime, String infoCode, String formula) throws Exception {
        ComputeCalculateDTO computeCalculateDTO = new ComputeCalculateDTO();
        computeCalculateDTO.setObjId(objId);
        computeCalculateDTO.setProjectId(projectDTO.getProjectId());
        computeCalculateDTO.setTargetId(infoCode);
        Map<String, Double> timeDataMap = TimeDataUtil.getTimeDataMap(lastComputeTime, computeEndTime,
                EnumTimeType.MIN15);

        Date timeEnd = null;
        for (Map.Entry<String, Double> entry : timeDataMap.entrySet()) {
            Date dataTime = DateUtils.str2Date(entry.getKey(), DateUtils.SDF_SECOND);
            timeEnd = dataTime;
            if (dataTime.getTime() > System.currentTimeMillis()) {
                break;
            }
            computeCalculateDTO.setDataTime(dataTime);
            Double targetData;
            targetData = computeObjTargetDataByCalculate(computeCalculateDTO, formula);
            if (targetData == null) {
                break;
            }

            MonitorIndicatorRecord monitorIndicatorRecord = new MonitorIndicatorRecord();
            monitorIndicatorRecord.setMonitorIndicatorId(infoCode);
            monitorIndicatorRecord.setDataTime(dataTime);
            monitorIndicatorRecord.setObjId(objId);
            monitorIndicatorRecord.setProject(projectDTO.getProjectId());
            monitorIndicatorRecord.setDataValue(targetData);
            saveOrUpdateByPk(monitorIndicatorRecord);
        }
        return timeEnd;
    }

    private Date getDefultStartTime() throws Exception {
        Date lastComputeTime;
        if (StringUtils.isBlank(computePropertiesConfig.getDStartTime())) {

            lastComputeTime = DateUtils.get15MinLite(new Date());
        } else {
            lastComputeTime = DateUtils.str2Date(computePropertiesConfig.getDStartTime(), DateUtils.SDF_SECOND);
        }
        return lastComputeTime;
    }


    private void saveOrUpdateComputeTime(String projectId, String objId, String indicatorId, Date computeTime) {
        if (computeTime == null) {
            return;
        }
        CalculateTimeRecord update = new CalculateTimeRecord();
        update.setMonitorIndicatorId(indicatorId);
        update.setObjId(objId);
        update.setProjectId(projectId);
        update.setCalculateLastTime(computeTime);
        calculateTimeRecordService.saveOrUpdateByPk(update);
    }



    @Override
    @Async("asyncSendDataExecutor")
    public Future<String> sendIndicatorData(ProjectDTO projectDTO, CountDownLatch latch) throws Exception {
        long time = System.currentTimeMillis() + (20 * 1000);
        try {
            if (!lockUtil.lock(projectDTO.getProjectId() + "_sendData", String.valueOf(time))) {
                // 获取不到锁,略过
                return new AsyncResult<>("【监测指标数据计算线程】:项目:" + projectDTO.getProjectId() + "被占用");
            }

            long start = System.currentTimeMillis();
            sendIndicatorDataToAlarmServer(projectDTO);
            long end = System.currentTimeMillis();
            return new AsyncResult<>("【监测指标数据发送线程】发送项目指标数据用时:" + projectDTO.getProjectId() + ";" + (end - start));
        } finally {
            lockUtil.unlock(projectDTO.getProjectId() + "_sendData", String.valueOf(time));
            latch.countDown();
        }
    }

    @Override
    public void saveOrUpdateByPk(MonitorIndicatorRecord monitorIndicatorRecord) throws Exception {
        MonitorIndicatorRecord query = new MonitorIndicatorRecord();
        query.setProject(monitorIndicatorRecord.getProject());
        query.setObjId(monitorIndicatorRecord.getObjId());
        query.setDataTime(monitorIndicatorRecord.getDataTime());
//        query.setSplitTimeType(EnumTimeType.MIN15.getCode());
        query.setMonitorIndicatorId(monitorIndicatorRecord.getMonitorIndicatorId());
        List<MonitorIndicatorRecord> queryList = coredao.query(query);
        if (CollectionUtils.isEmpty(queryList)) {//新增
            coredao.save(monitorIndicatorRecord);
        } else {
            MonitorIndicatorRecord update = new MonitorIndicatorRecord();
            update.setDataValue(monitorIndicatorRecord.getDataValue());
            coredao.update(query, update);
        }
    }

    @Override
    public void cleanAlarmConfig(String projectId) {
        if (StringUtils.isBlank(projectId)) {
            redisUtil.del(DiAgnoseConst.RedisConstants.PROJECT_ALARM_ITEM);
        } else {
            redisUtil.hdel(DiAgnoseConst.RedisConstants.PROJECT_ALARM_ITEM, projectId);
        }
    }

    @Override
    public void cleanAlarmConfigSendTime(CleanAlarmConfigDTO cleanAlarmConfigDTO) {
        if (StringUtils.isBlank(cleanAlarmConfigDTO.getProjectId())) {
            //清除所有
            redisUtil.del(DiAgnoseConst.RedisConstants.PROJECT_ITEM_SEND_TIME);
        } else if (StringUtils.isBlank(cleanAlarmConfigDTO.getAlarmConfigItemId())) {
            //清除项目
            redisUtil.del(DiAgnoseConst.RedisConstants.PROJECT_ALARM_ITEM + ":" + cleanAlarmConfigDTO.getProjectId());
        } else {
            //清除指定
            redisUtil.hdel(DiAgnoseConst.RedisConstants.PROJECT_ALARM_ITEM + ":" + cleanAlarmConfigDTO.getProjectId(), cleanAlarmConfigDTO.getAlarmConfigItemId());
        }
    }

    private void sendIndicatorDataToAlarmServer(ProjectDTO projectDTO) throws Exception {
        //通过报警服务查询报警规则
        List<AlarmRuleDTO> projectAlarmRules = energyAlarmService.getProjectAlarmRules(projectDTO.getProjectId());
        if (CollectionUtils.isEmpty(projectAlarmRules)) {
            return;
        }
        //根据报警规则,获取对象id_报警类型及设备监控点位list
        Map<String, List<String>> objIdAlarmItemCodeAndInfoCodes = getObjIdAlarmItemCodeAndInfoCodes(projectAlarmRules);
        if (CollectionUtils.isEmpty(objIdAlarmItemCodeAndInfoCodes) || objIdAlarmItemCodeAndInfoCodes.size() == 0){
            return;
        }
        for (Map.Entry<String, List<String>> entry : objIdAlarmItemCodeAndInfoCodes.entrySet()) {
            String objIdAndAlarmItemCode = entry.getKey();
            int indexOf = objIdAndAlarmItemCode.indexOf("_");
            String objId = objIdAndAlarmItemCode.substring(0, indexOf);
            String alarmItemCode = objIdAndAlarmItemCode.substring(indexOf + 1);
            String lockKey = objId+"-"+alarmItemCode;
            Object isLock = redisUtil.get(lockKey);
            if (null == isLock){
                continue;
            }
            long time = System.currentTimeMillis() + (10 * 1000);
            lockUtil.sendDataLock(lockKey, String.valueOf(time));

            List<String> infCodeList = entry.getValue();

            //通过redis查询对象该报警类型下的发数截止时间
            ConfigItemDataSendTimeDTO configItemSendTimeDTO = getConfigItemSendTimeDTO(projectDTO, objIdAndAlarmItemCode);

            Date lastSendTime = configItemSendTimeDTO.getLastSendTime();
            Date endTime = getSendEndTime(lastSendTime);

            Map<String, List<MonitorIndicatorRecord>> timeKeyAndRecordList = getTimekeyAndRecordListGteLt(projectDTO, objId, infCodeList, lastSendTime, endTime);
            if (timeKeyAndRecordList == null) {
                continue;
            }
           Map<String, List<MonitorIndicatorRecord>> sendTimeKeyAndDataList = new LinkedHashMap<>();
            for (Map.Entry<String, List<MonitorIndicatorRecord>> dateListEntry : timeKeyAndRecordList.entrySet()) {
                List<MonitorIndicatorRecord> value = dateListEntry.getValue();
                if (CollectionUtils.isEmpty(value) || value.size() != infCodeList.size()) {
                    continue;
                }
                sendTimeKeyAndDataList.put(dateListEntry.getKey(),value);
                endTime = DateUtils.str2Date(dateListEntry.getKey(), DateUtils.SDFSECOND);
            }
            JSONArray sendArray = CollectDataUtil.batchBuildSendJsonParam(sendTimeKeyAndDataList,alarmItemCode);
            rabbitConfig.sendMsg(sendArray);
//          String sentValue = CollectDataUtil.batchBuildSendParam(sendTimeKeyAndDataList,alarmItemCode);
            //AlarmWebSocketServer.sendMsgToClients(projectDTO.getProjectId(), sentValue);
            log.error("指标发送报表服务数据:" + projectDTO.getProjectId()+";"+ sendArray);
            configItemSendTimeDTO.setLastSendTime(endTime);
            configItemSendTimeDTO.setUpdateTime(new Date());
            updateRedisConfigItemSendTimeDTO(projectDTO.getProjectId(), objIdAndAlarmItemCode, configItemSendTimeDTO);
        }
    }

    /**
     * @title getTimekeyAndRecordListGteLt
     * @description 查询该时间段的需要发送的指标数据,key为时间字符串
     * @author leifeng
     * @param: projectDTO
     * @param: objId
     * @param: infCodeList
     * @param: lastSendTime
     * @param: endTime
     * @updateTime 2021/11/5 10:59
     * @return: java.util.Map<java.lang.String,java.util.List<com.persagy.framework.ems.data.pojo.hbase.MonitorIndicatorRecord>>
     * @throws
     */
    private Map<String, List<MonitorIndicatorRecord>> getTimekeyAndRecordListGteLt(ProjectDTO projectDTO, String objId, List<String> infCodeList, Date lastSendTime, Date endTime) throws Exception {
        List<MonitorIndicatorRecord> list = getMonitorIndicatorRecordsGtLte(projectDTO.getProjectId(), objId, lastSendTime, infCodeList, endTime);
        if (CollectionUtils.isEmpty(list)) {
            return null;
        }
        Map<String, List<MonitorIndicatorRecord>> timeKeyAndRecordList = new LinkedHashMap<>();
        for (MonitorIndicatorRecord record : list) {
            String timeKey = DateUtils.date2Str(record.getDataTime(), DateUtils.SDFSECOND);
            if (!timeKeyAndRecordList.containsKey(timeKey)) {
                timeKeyAndRecordList.put(timeKey, new ArrayList<>());
            }
            timeKeyAndRecordList.get(timeKey).add(record);
        }
        return timeKeyAndRecordList;
    }

    /**
     * @title getObjIdAlarmItemCodeAndInfoCodes
     * @description 获取对象报警类型及监控指标list,key为objId_itemCode
     * @author leifeng
     * @param: projectAlarmRules
     * @updateTime 2021/10/26 21:35
     * @return: java.util.Map<java.lang.String,java.util.List<java.lang.String>>
     * @throws
     */
    private Map<String, List<String>> getObjIdAlarmItemCodeAndInfoCodes(List<AlarmRuleDTO> projectAlarmRules) {
        Map<String, List<String>> objIdAlarmItemCodeAndInfoCodes = new HashMap<>();
        for (AlarmRuleDTO projectAlarmRule : projectAlarmRules) {
            String alarmItemCode = projectAlarmRule.getAlarmItemCode();
            List<String> objIds = projectAlarmRule.getObjIds();
            List<String> infoCodes = projectAlarmRule.getInfoCodes();
            if (!CollectionUtils.isEmpty(infoCodes) && !CollectionUtils.isEmpty(objIds)){
                for (String infoCode : infoCodes) {
                    for (String objId : objIds) {
                        String key = objId + "_" + alarmItemCode;
                        if (!objIdAlarmItemCodeAndInfoCodes.containsKey(key)) {
                            objIdAlarmItemCodeAndInfoCodes.put(key, new ArrayList<String>());
                        }
                        objIdAlarmItemCodeAndInfoCodes.get(key).add(infoCode);
                    }
                }
            }
        }
        return objIdAlarmItemCodeAndInfoCodes;
    }

    private Date getSendEndTime(Date lastSendTime) {
        Date date = DateUtils.addHours(lastSendTime, computePropertiesConfig.getSendDataHour());
        if (date.getTime() > System.currentTimeMillis()) {//大于当前时间
            return new Date();
        }
        return date;
    }

    /**
     * @title getMonitorIndicatorRecordsGtLte
     * @description 根据项目, 对象id, 指标id, 数据时间查询指标数据list
     * @author leifeng
     * @param: projectDTO
     * @param: alarmConfigItem
     * @param: lastSendTime
     * @param: infoCodes
     * @param: endTime
     * @updateTime 2021/9/25 20:08
     * @return: java.util.List<com.persagy.apm.diagnose.indicatorrecord.model.MonitorIndicatorRecord>
     */
    private List<MonitorIndicatorRecord> getMonitorIndicatorRecordsGtLte(String projectId, String objId, Date lastSendTime, List<String> infoCodes, Date endTime) throws Exception {
        MonitorIndicatorRecord monitorIndicatorRecord = new MonitorIndicatorRecord();
        monitorIndicatorRecord.setProject(projectId);
        monitorIndicatorRecord.setObjId(objId);
//        monitorIndicatorRecord.setSplitTimeType(EnumTimeType.MIN15.getCode());
        monitorIndicatorRecord.setSpecialOperation("monitorIndicatorId", SpecialOperator.$in, infoCodes);
        monitorIndicatorRecord.setSpecialOperation("dataTime", SpecialOperator.$gt, lastSendTime);
        monitorIndicatorRecord.setSpecialOperation("dataTime", SpecialOperator.$lte, endTime);
        monitorIndicatorRecord.setSort("dataTime", EMSOrder.Asc);
        return coredao.query(monitorIndicatorRecord);
    }


    /**
     * @title getConfigItemSendTimeDTO
     * @description redis中查询报警条件上次发数时间, 如果查询为空, 则初始化当前时间
     * @author leifeng
     * @param: projectDTO
     * @param: alarmConfigItem
     * @updateTime 2021/9/25 20:03
     * @return: com.persagy.apm.diagnose.indicatorrecord.model.dto.ConfigItemDataSendTimeDTO
     */
    private ConfigItemDataSendTimeDTO getConfigItemSendTimeDTO(ProjectDTO projectDTO, String objIdAndAlarmItemCode) throws Exception {
        ConfigItemDataSendTimeDTO sendTimeDTO = (ConfigItemDataSendTimeDTO) redisUtil.get(getRedisSendDataKey(projectDTO.getProjectId(), objIdAndAlarmItemCode));
        if (sendTimeDTO == null) {
            sendTimeDTO = ConfigItemDataSendTimeDTO.builder().lastSendTime(getDefultStartTime()).build();
        }
        return sendTimeDTO;
    }

    /**
     * @throws
     * @title getRedisSendDataKey
     * @description 获取发数时间redis key
     * @author leifeng
     * @param: projectId
     * @param: alarmConfigItemId
     * @updateTime 2021/10/14 19:56
     * @return: java.lang.String
     */
    private String getRedisSendDataKey(String projectId, String objIdAndAlarmItemCode) {
        return DiAgnoseConst.RedisConstants.PROJECT_ITEM_SEND_TIME + ":" + projectId + ":" + objIdAndAlarmItemCode;
    }

    /**
     * @title updateRedisConfigItemSendTimeDTO
     * @description 更新缓存中报警配置发数时间
     * @author leifeng
     * @param: projectId
     * @param: sendTimeDTO
     * @updateTime 2021/9/25 20:05
     */
    private void updateRedisConfigItemSendTimeDTO(String projectId, String objeIdAndAlarmItemCode, ConfigItemDataSendTimeDTO sendTimeDTO) {
        redisUtil.set(getRedisSendDataKey(projectId, objeIdAndAlarmItemCode), sendTimeDTO);
    }

    /**
     * @title updateProjectComputeRound
     * @description 更新项目计算轮次
     * @author leifeng
     * @param: projectId
     * @param: projectCompute
     * @updateTime 2021/9/25 20:08
     */
    private void updateProjectComputeRound(String projectId, ProjectIndicatorCompute projectCompute) {
        Integer round = 1;
        ProjectIndicatorCompute projectIndicatorCompute = new ProjectIndicatorCompute();
        projectIndicatorCompute.setProjectId(projectId);

        projectIndicatorCompute.setComputeTime(new Date());
        if (projectCompute == null) {
            projectIndicatorCompute.setComputeRound(round);
            projectIndicatorComputeService.save(projectIndicatorCompute);
        } else {
            round = getComputeRound(projectCompute.getComputeRound());
            projectIndicatorCompute.setComputeRound(round);
            projectIndicatorComputeService.updateById(projectIndicatorCompute);
        }
        redisUtil.hset(DiAgnoseConst.RedisConstants.TargetComputeTime, projectId, round.toString());
    }


    /**
     * @title getComputeRound
     * @description 获取计算轮次
     * @author leifeng
     * @param: computeRound
     * @updateTime 2021/9/25 20:09
     * @return: java.lang.Integer
     */
    private Integer getComputeRound(Integer computeRound) {
        if (computeRound > 10000000) {
            return 1;
        }
        return ++computeRound;
    }



    /**
     * @ClassName MonitorIndicatorRecordServiceImpl
     * @author leifeng
     * @Description 查询项目下对象指标计算时间,key为objId_indicator
     * @createTime 2021/9/24 11:22
     */
    private Map<String, Date> getObjId_IndicatorIdAndComputeTIme(ProjectDTO projectDTO) {
        List<CalculateTimeRecord> calculateTimeRecords = calculateTimeRecordService.queryTimeRecordListByProjectId(projectDTO.getProjectId());
        return calculateTimeRecords.stream().collect(Collectors.toMap(a -> a.getObjId() + alarmSlipt + a.getMonitorIndicatorId(), CalculateTimeRecord::getCalculateLastTime));
    }



    private Double computeObjTargetDataByCalculate(ComputeCalculateDTO computeCalculateDTO, String calculateStr)
            throws Exception {
        List<String> variables = CalculateUtil.getVariablesFromFormula(calculateStr);
        Map<String, Double> variableAndData = new HashMap<>(16);
        for (String variable : variables) {
            // 获取${}中变量
            String v = variable.substring(2, variable.length() - 1);
            // 解析变量
            Double data = null;

            if (v.startsWith(CalculateUtil.SUM)) {
                // 求和
                v = v.substring(4, v.length() - 1);
                computeCalculateDTO.setCalculateVariableDTO(CalculateVariableDTO.buildCalculateVariableDTO(v));
                List<Double> dataList = analysisVariable(computeCalculateDTO);
                data = getSumDataFromDataList(dataList);
            } else if (v.startsWith(CalculateUtil.MIN)) {
                // 求最小值
                v = v.substring(4, v.length() - 1);
                computeCalculateDTO.setCalculateVariableDTO(CalculateVariableDTO.buildCalculateVariableDTO(v));
                List<Double> analysisVariable = analysisVariable(computeCalculateDTO);

                data = getMinDataFromDataList(analysisVariable);
            } else {
                computeCalculateDTO.setCalculateVariableDTO(CalculateVariableDTO.buildCalculateVariableDTO(v));
                List<Double> analysisVariable = analysisVariable(computeCalculateDTO);
                if (!CollectionUtils.isEmpty(analysisVariable)) {
                    data = analysisVariable.get(0);
                }
            }
            variableAndData.put(variable, data);
            if (data == null) {
                break;
            }
        }
        for (Map.Entry<String, Double> entry : variableAndData.entrySet()) {
            if (entry.getValue() == null) {
                return null;
            }
            calculateStr = calculateStr.replace(entry.getKey(), variableAndData.get(entry.getKey()).toString());
        }
        BigDecimal result = new ComputeFormulaUtil(calculateStr).getResult();
        if (result == null) {
            log.error("【监测指标数据计算线程】:解析公式为空,projectId:" + computeCalculateDTO.getProjectId() + "对象id:"
                    + computeCalculateDTO.getObjId() + ";指标id:" + computeCalculateDTO.getTargetId()
                    + ";带入公式:" + calculateStr);
            return null;
        }
        return result.doubleValue();
    }

    private Double getMinDataFromDataList(List<Double> analysisVariable) {
        if (CollectionUtils.isEmpty(analysisVariable)) {
            return null;
        }
        Double data = null;
        for (Double double1 : analysisVariable) {
            if (double1 == null) {
                continue;
            }
            if (double1 == 0.0) {
                continue;
            }
            data = data == null ? double1 : data > double1 ? double1 : data;
        }
        return data;
    }

    private Double getSumDataFromDataList(List<Double> dataList) {
        double data = 0.0;
        if (CollectionUtils.isEmpty(dataList)) {
            return null;
        }
        for (Double double1 : dataList) {
            if (double1 == null) {
                return null;
            }
            data += double1;
        }
        return data;
    }

    private List<Double> analysisVariable(ComputeCalculateDTO computeCalculateDTO) throws Exception {
        CalculateVariableDTO calculateVariableDTO = computeCalculateDTO.getCalculateVariableDTO();
        if (calculateVariableDTO.getComputeObjType().equals(CalculateUtil.OBJTYPE)) {
            // 当前对象
            return computeObjDataByComputeCalculateDTO(computeCalculateDTO);
        } else {
            List<JSONObject> objList = centerDataService.queryObjRelationList(CenterDataUtil.getObjRelationDataDTO(
                    computeCalculateDTO.getProjectId(), calculateVariableDTO.getComputeObjType(),
                    computeCalculateDTO.getObjId(), null));
            if (CollectionUtils.isEmpty(objList)) {
                log.error("【指标计算线程】:查询数据中台子对象为空,projectId:" + computeCalculateDTO.getProjectId() + ";objId:"
                        + computeCalculateDTO.getObjId() + ";子对象类型:"
                        + calculateVariableDTO.getComputeObjType());
                return null;
            }
            List<Double> list = new ArrayList<>();
            for (JSONObject object : objList) {
                ComputeCalculateDTO computeCalculateDTO2 = new ComputeCalculateDTO();
                BeanUtils.copyProperties(computeCalculateDTO2, computeCalculateDTO);
                computeCalculateDTO2.setObjId(object.getString("id"));
                List<Double> queryDataList = computeObjDataByComputeCalculateDTO(computeCalculateDTO2);
                list.addAll(queryDataList);
            }
            return list;
        }
    }

    private List<Double> computeObjDataByComputeCalculateDTO(ComputeCalculateDTO computeCalculateDTO) throws
            Exception {
        CalculateVariableDTO calculateVariableDTO = computeCalculateDTO.getCalculateVariableDTO();
        String computeCode = calculateVariableDTO.getComputeCode();
        String computeType = calculateVariableDTO.getComputeType();
        String time = calculateVariableDTO.getTime();
        String timeType = calculateVariableDTO.getTimeType();
        Date startTime = CalculateUtil.getStartTime(time, computeCalculateDTO.getDataTime());
        if (computeType.equals(CalculateUtil.InfoCode)) {
            // 查询数据中台信息点数据
            if (time.equals(CalculateUtil.invalidTime)) {
                // 查询静态信息点
                return getObjStaticDataFromCenterData(computeCalculateDTO.getProjectId(), computeCalculateDTO.getObjId(), computeCode);
            } else {
                assert startTime != null;
                return getServiceDataFromCenterDataService(computeCalculateDTO.getProjectId(),
                        computeCalculateDTO.getObjId(), computeCode, timeType, startTime,
                        DateUtils.addHours(startTime, computePropertiesConfig.getComputeAfterHour()));
            }
        } else {
            // 查询指标数据
            List<Double> result = new ArrayList<>();
            MonitorIndicatorRecord query = new MonitorIndicatorRecord();
            query.setProject(computeCalculateDTO.getProjectId());
            query.setObjId(computeCalculateDTO.getObjId());
            query.setDataTime(startTime);
            query.setMonitorIndicatorId(computeCode);
            List<MonitorIndicatorRecord> list = coredao.query(query);
            if (CollectionUtils.isEmpty(list)) {
                return result;
            }
            result.add(list.get(0).getDataValue());
            return result;
        }
    }

    private List<Double> getObjStaticDataFromCenterData(String projectId, String objId, String computeCode) throws Exception {
        List<Double> arrayList = new ArrayList<>();
        List<JSONObject> objectList = centerDataService.queryObjListByObjId(projectId, Arrays.asList(objId));
        if (CollectionUtils.isEmpty(objectList)) {
            log.error("【指标数据计算线程】:查询数据中台对象不存在,:objId:" + objId + ";信息点:" + computeCode);
            arrayList.add(null);
            return arrayList;
        }
        Double double1 = objectList.get(0).getDouble(computeCode);
        if (double1 == null) {
            log.error("【指标数据计算线程】:查询数据中台对象静态数据为空,:objId:" + objId + ";信息点:" + computeCode);
        }
        arrayList.add(double1);
        return arrayList;
    }

    private List<Double> getServiceDataFromCenterDataService(String projectId, String objId, String computeCode,
                                                             String timeType, Date startTime, Date startTime2) throws Exception {
        List<TimeDataDTO> dataList = centerDataService.queryObjTimeDataList(projectId, objId, computeCode, startTime,
                startTime2, EnumDataTimeType.getEnumDataTimeType(timeType));
        List<Double> result = new ArrayList<>();
        if (!CollectionUtils.isEmpty(dataList)) {
            result.add(dataList.get(0).getData());
        }
        if (CollectionUtils.isEmpty(result)) {
            log.error("【指标计算线程】:查询数据中台对象历史数据为空,projectId:" + projectId + ";objId:" + objId + ";信息点:" + computeCode
                    + ";时间:" + DateUtils.date2Str(startTime, DateUtils.SDF_SECOND) + "_"
                    + DateUtils.date2Str(startTime2, DateUtils.SDF_SECOND));
        }
        return result;
    }
}
